@@ -1,6 +1,6 @@
 /* sedcomp.c -- stream editor main and compilation phase
    Copyright (C) 1995-2003 Eric S. Raymond
-   Copyright (C) 2004-2006 Rene Rebe
+   Copyright (C) 2004-2021 Rene Rebe
 
    The stream editor compiles its command input  (from files or -e options)
 into an internal form using compile() then executes the compiled form using
@@ -32,11 +32,11 @@ sedcmd	cmds[MAXCMDS+1];	/* hold compiled commands */
 long	linenum[MAXLINES];	/* numeric-addresses table */
 
 /* miscellaneous shared variables */ 
-int	nflag;			/* -n option flag */
+short	nflag;			/* -n option flag */
 int	eargc;			/* scratch copy of argument count */
 sedcmd	*pending	= NULL;	/* next command to be executed */
 
-int	last_line_used = 0;	/* last line address ($) was used */
+short	last_line_used = 0;	/* last line address ($) was used */
 
 void die (const char* msg) {
 	fprintf(stderr, "sed: ");
@@ -57,32 +57,31 @@ void die (const char* msg) {
 #define IFEQ(x, v)	if (*x == v) x++ , /* do expression */
 
 /* error messages */
-static char	AGMSG[]	= "garbled address %s";
-static char	CGMSG[]	= "garbled command %s";
-static char	TMTXT[]	= "too much text: %s";
-static char	AD1NG[]	= "no addresses allowed for %s";
-static char	AD2NG[]	= "only one address allowed for %s";
-static char	TMCDS[]	= "too many commands, last was %s";
-static char	COCFI[]	= "cannot open command-file %s";
-static char	UFLAG[]	= "unknown flag %c";
-/*static char	COOFI[]	= "cannot open %s";*/
-static char	CCOFI[]	= "cannot create %s";
-static char	ULABL[]	= "undefined label %s";
-static char	TMLBR[]	= "too many {'s";
-static char	FRENL[]	= "first RE must be non-null";
-static char	NSCAX[]	= "no such command as %s";
-static char	TMRBR[]	= "too many }'s";
-static char	DLABL[]	= "duplicate label %s";
-static char	TMLAB[]	= "too many labels: %s";
-static char	TMWFI[]	= "too many w files";
-static char	REITL[]	= "RE too long: %s";
-static char	TMLNR[]	= "too many line numbers";
-static char	TRAIL[]	= "command \"%s\" has trailing garbage";
-static char	RETER[] = "RE not terminated: %s";
-static char	CCERR[] = "unknown character class: %s";
+static const char	AGMSG[]	= "garbled address %s";
+static const char	CGMSG[]	= "garbled command %s";
+static const char	TMTXT[]	= "too much text: %s";
+static const char	AD1NG[]	= "no addresses allowed for %s";
+static const char	AD2NG[]	= "only one address allowed for %s";
+static const char	TMCDS[]	= "too many commands, last was %s";
+static const char	COCFI[]	= "cannot open command-file %s";
+static const char	UFLAG[]	= "unknown flag %c";
+static const char	CCOFI[]	= "cannot create %s\n";
+static const char	ULABL[]	= "undefined label %s\n";
+static const char	TMLBR[]	= "too many {'s";
+static const char	FRENL[]	= "first RE must be non-null";
+static const char	NSCAX[]	= "no such command as %s";
+static const char	TMRBR[]	= "too many }'s";
+static const char	DLABL[]	= "duplicate label %s";
+static const char	TMLAB[]	= "too many labels: %s";
+static const char	TMWFI[]	= "too many w files";
+static const char	REITL[]	= "RE too long: %s";
+static const char	TMLNR[]	= "too many line numbers";
+static const char	TRAIL[]	= "command \"%s\" has trailing garbage";
+static const char	RETER[] = "RE not terminated: %s";
+static const char	CCERR[] = "unknown character class: %s";
 
 /* cclass to c function mapping ,-) */
-const char* cclasses[] = {
+static const char* cclasses[] = {
 	"alnum", "a-zA-Z0-9",
 	"lower", "a-z",
 	"space", " \f\n\r\t\v",
@@ -95,7 +94,7 @@ const char* cclasses[] = {
 	"print", " -\x7e",
 	"graph", "!-\x7e",
 	"punct", "!-/:-@[-`{-\x7e",
-	NULL, NULL};
+	NULL};
  
 typedef struct			/* represent a command label */
 {
@@ -116,7 +115,7 @@ static char	*poolend = pool + POOLSIZE;	/* pointer past pool end */
 
 /* compilation state */
 static FILE	*cmdf	= NULL;		/* current command source */
-static char	*cp	= linebuf;	/* compile pointer */
+static char	*cp;			/* compile pointer */
 static sedcmd	*cmdp	= cmds;		/* current compiled-cmd ptr */
 static char	*lastre	= NULL;		/* old RE pointer */
 static int	bdepth	= 0;		/* current {}-nesting level */
@@ -124,8 +123,8 @@ static int	bcount	= 0;		/* # tagged patterns in current RE */
 static char	**eargv;		/* scratch copy of argument list */
 
 /* compilation flags */
-static int	eflag;			/* -e option flag */
-static int	gflag;			/* -g option flag */
+static short	eflag;			/* -e option flag */
+static short	gflag;			/* -g option flag */
 
 /* prototypes */
 static char *address(char *expbuf);
@@ -149,32 +148,32 @@ int main(int argc, char *argv[])
 	eargv	= argv;		/* set local copy of argument list */
 	cmdp->addr1 = pool;	/* 1st addr expand will be at pool start */
 	if (eargc == 1)
-		exit(0);	/* exit immediately if no arguments */
+		return 0;	/* exit immediately if no arguments */
 
 	/* scan through the arguments, interpreting each one */
 	while ((--eargc > 0) && (**++eargv == '-'))
 		switch (eargv[0][1])
 		{
 		case 'e':
-			eflag++; compile();	/* compile with e flag on */
+			eflag = 1; compile();	/* compile with e flag on */
 			eflag = 0;
 			continue;		/* get another argument */
 		case 'f':
 			if (eargc-- <= 0)	/* barf if no -f file */
-				exit(2);
+				return 2;
 			if ((cmdf = fopen(*++eargv, "r")) == NULL)
 			{
 				fprintf(stderr, COCFI, *eargv);
-				exit(2);
+				return 2;
 			}
 			compile();	/* file is O.K., compile it */
 			fclose(cmdf);
 			continue;	/* go back for another argument */
 		case 'g':
-			gflag++;	/* set global flag on all s cmds */
+			gflag = 1;	/* set global flag on all s cmds */
 			continue;
 		case 'n':
-			nflag++;	/* no print except on p flag or w */
+			nflag = 1;	/* no print except on p flag or w */
 			continue;
 		default:
 			fprintf(stdout, UFLAG, eargv[0][1]);
@@ -184,7 +183,7 @@ int main(int argc, char *argv[])
 	if (cmdp == cmds)	/* no commands have been compiled */
 	{
 		eargv--; eargc++;
-		eflag++; compile(); eflag = 0;
+		eflag = 1; compile(); eflag = 0;
 		eargv++; eargc--;
 	}
 
@@ -197,7 +196,7 @@ int main(int argc, char *argv[])
 		execute(NULL);	/*   execute commands from stdin only */
 	else while(--eargc>=0)	/* else execute only -e commands */
 		execute(*eargv++);
-	exit(0);		/* everything was O.K. if we got here */
+	return 0;		/* everything was O.K. if we got here */
 }
 
 #define	H	0x80	/* 128 bit, on if there's really code for command */
@@ -221,6 +220,11 @@ static char	cmdmask[] =
 static void compile(void)
 {
 	char	ccode;
+	if (cmdline(cp = linebuf) < 0)		/* start parsing a new line */
+		return;
+	/* no commands have been compiled? #n handling */
+	if (cmdp == cmds && *cp == '#' && cp[1] == 'n')
+		nflag = 1;	/* no print except on p flag or w */
 
 	for(;;)					/* main compilation loop */
 	{
@@ -333,13 +337,13 @@ static int cmdcomp(char cchar)
 		cmpstk[bdepth++] = &(cmdp->u.link);
 		if (++cmdp >= cmds + MAXCMDS) die(TMCDS);
 		if (*cp == '\0') *cp++ = ';', *cp = '\0';	/* get next cmd w/o lineread */
-		return(1);
+		return 1;
 
 	case '}':	/* end command group */
 		if (cmdp->addr1) die(AD1NG);	/* no addresses allowed */
 		if (--bdepth < 0) die(TMRBR);	/* too many right braces */
 		*cmpstk[bdepth] = cmdp;		/* set the jump address */
-		return(1);
+		return 1;
 
 	case '=':			/* print current source line number */
 	case 'q':			/* exit the stream editor */
@@ -360,7 +364,7 @@ static int cmdcomp(char cchar)
 			if (++lab >= labels + MAXLABS) die(TMLAB);
 		}
 		lpt->address = cmdp;
-		return(1);
+		return 1;
 
 	case 'b':	/* branch command */
 	case 't':	/* branch-on-succeed command */
@@ -441,7 +445,7 @@ static int cmdcomp(char cchar)
 			IFEQ(cp, 'g') cmdp->flags.global++;
 			IFEQ(cp, 'p') cmdp->flags.print = 1;
 			IFEQ(cp, 'P') cmdp->flags.print = 2;
-			if(isdigit(*cp))
+			if (isdigit(*cp))
 			{
 				if (cmdp->nth)
 					break; /* no multiple n args */
@@ -467,13 +471,13 @@ static int cmdcomp(char cchar)
 			if (strcmp(fname[nwfiles], fname[i]) == 0)
 			{
 				cmdp->fout = fout[i];
-				return(0);
+				return 0;
 			}
 		/* if didn't find one, open new out file */
 		if ((cmdp->fout = fopen(fname[nwfiles], "w")) == NULL)
 		{
 			fprintf(stderr, CCOFI, fname[nwfiles]);
-			exit(2);
+			return 2;
 		}
 		fout[nwfiles++] = cmdp->fout;
 		break;
@@ -484,7 +488,7 @@ static int cmdcomp(char cchar)
 		if (fp > poolend) die(TMTXT);		/* fail on overflow */
 		break;
 	}
-	return(0);	/* succeeded in interpreting one command */
+	return 0;	/* succeeded in interpreting one command */
 }
 
 /* generate replacement string for substitute command right hand side
@@ -504,7 +508,7 @@ static char *rhscomp(char* rhsp, char delim)	/* uses bcount */
 				*rhsp = *p++;
 				/* check validity of pattern tag */
 				if (*rhsp > bcount + '0')
-					return(BAD);
+					return BAD;
 				*rhsp++ |= 0x80; /* mark the good ones */
 			}
 			else /* escape */
@@ -522,7 +526,7 @@ static char *rhscomp(char* rhsp, char delim)	/* uses bcount */
 		{
 			*rhsp++ = '\0';		/* cap the expression string */
 			cp = p;
-			return(rhsp);		/* pt at 1 past the RE */
+			return rhsp;		/* pt at 1 past the RE */
 		}
 		else if (*rhsp == '&')		/* special case, convert to backref \0 */
 		{
@@ -530,7 +534,7 @@ static char *rhscomp(char* rhsp, char delim)	/* uses bcount */
 			goto dobackref;
 		}
 		else if (*rhsp++ == '\0')	/* last ch not RE end, help! */
-			return(BAD);
+			return BAD;
 }
 
 /* compile a regular expression to internal form
@@ -552,7 +556,7 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 	int		tags;		/* # of closed tags */
 
 	if (*cp == redelim) {		/* if first char is RE endmarker */
-	    return(ep);
+	    return ep;
 	}
 
 	lastep = lastep2 = NULL;	/* there's no previous RE */
@@ -567,14 +571,14 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 		if (*sp == 0) /* no termination */
 			die (RETER);
 		if (ep >= expbuf + RELIMIT)	/* match is too large */
-			return(cp = sp, BAD);
+			return cp = sp, BAD;
 		if ((c = *sp++) == redelim)	/* found the end of the RE */
 		{
 			cp = sp;
 			if (brnestp != brnest)	/* \(, \) unbalanced */
-				return(BAD);
+				return BAD;
 			*ep++ = CEOF;		/* write end-of-pattern mark */
-			return(ep);		/* return ptr to compiled RE */
+			return ep;		/* return ptr to compiled RE */
 		}
 
 		lastep = lastep2;
@@ -586,7 +590,7 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 			if ((c = *sp++) == '(')	/* start tagged section */
 			{
 				if (bcount >= MAXTAGS)
-					return(cp = sp, BAD);
+					return cp = sp, BAD;
 				*brnestp++ = bcount;	/* update tag stack */
 				*ep++ = CBRA;		/* enter tag-start */
 				*ep++ = bcount++;	/* bump tag count */
@@ -596,7 +600,7 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 			else if (c == ')')	/* end tagged section */
 			{
 				if (brnestp <= brnest)	/* extra \) */
-					return(cp = sp, BAD);
+					return cp = sp, BAD;
 				*ep++ = CKET;		/* enter end-of-tag */
 				*ep++ = *--brnestp;	/* pop tag stack */
 				tags++;			/* count closed tags */
@@ -607,13 +611,13 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 			else if (c >= '1' && c <= '9' && c != redelim)	/* tag use, if !delim */
 			{
 				if ((c -= '1') >= tags)	/* too few */
-					return(BAD);
+					return BAD;
 				*ep++ = CBACK;		/* enter tag mark */
 				*ep++ = c;		/* and the number */
 				continue;
 			}
 			else if (c == '\n')	/* escaped newline no good */
-				return(cp = sp, BAD);
+				return cp = sp, BAD;
 			else if (c == 'n')		/* match a newline */
 				c = '\n';
 			else if (c == 't')		/* match a tab */
@@ -637,7 +641,7 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 			continue;
 
 		case '\n':	/* trailing pattern delimiter is missing */
-			return(cp = sp, BAD);
+			return cp = sp, BAD;
 
 		case '.':	/* match any char except newline */
 			*ep++ = CDOT;
@@ -657,7 +661,7 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 			continue;
 
 		case '[':	/* begin character set pattern */
-			if (ep + 17 >= expbuf + RELIMIT)
+			if (ep + 33 >= expbuf + RELIMIT)
 				die(REITL);
 			*ep++ = CCL;		/* insert class mark */
 			if ((negclass = ((c = *sp++) == '^')))
@@ -680,14 +684,14 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 					  const char **it;
 					  p2 = sp+1;
 					  for (p2 = sp+1;
-					       p2 < p-2 && p2-sp-1 < sizeof(cc);
+					       p2 < p-2 && p2-sp-1 < sizeof(cc) - 1;
 					       p2++)
 					    cc[p2-sp-1] = *p2;
 					  cc[p2-sp-1] = 0; /* termination */
 
 					  it = cclasses;
 					  while (*it && strcmp(*it, cc))
-						it +=2;
+						it += 2;
 					  if (!*it++)
 					    die(CCERR);
 
@@ -726,16 +730,16 @@ static char *recomp(char *expbuf, char redelim)	/* uses cp, bcount */
 
 				/* enter (possibly translated) char in set */
 				if (c)
-					ep[c >> 3] |= bits(c & 7);
+				  ep[((unsigned char)c) >> 3] |= bits(c & 7);
 			} while
 				((c = *sp++) != ']');
 
 			/* invert the bitmask if all-but was specified */
 			if (negclass)
-				for(classct = 0; classct < 16; classct++)
+				for(classct = 0; classct < 32; classct++)
 					ep[classct] ^= 0xFF;
 			ep[0] &= 0xFE;		/* never match ASCII 0 */ 
-			ep += 16;		/* advance ep past set mask */
+			ep += 32;		/* advance ep past set mask */
 			continue;
 
 		defchar:	/* match literal character */
@@ -770,40 +774,40 @@ static int cmdline(char	*cbuf)		/* uses eflag, eargc, cmdf */
 			while((*++cbuf = *p++))
 				if (*cbuf == '\\')
 				{
-					if ((*++cbuf = *p++) == '\0')
-						return(savep = NULL, -1);
+					if ((*++cbuf = *p++) == '0')
+						return savep = NULL, -1;
 					else
 						continue;
 				}
 				else if (*cbuf == '\n')	/* end of 1 cmd line */
 				{ 
 					*cbuf = '\0';
-					return(savep = p, 1);
+					return savep = p, 1;
 					/* we'll be back for the rest... */
 				}
 
 			/* found end-of-string; can advance to next argument */
-			return(savep = NULL, 1);
+			return savep = NULL, 1;
 		}
 
 		if ((p = savep) == NULL)
-			return(-1);
+			return -1;
 
 		while((*++cbuf = *p++))
 			if (*cbuf == '\\')
 			{
 				if ((*++cbuf = *p++) == '0')
-					return(savep = NULL, -1);
+					return savep = NULL, -1;
 				else
 					continue;
 			}
 			else if (*cbuf == '\n')
 			{
 				*cbuf = '\0';
-				return(savep = p, 1);
+				return savep = p, 1;
 			}
 
-		return(savep = NULL, 1);
+		return savep = NULL, 1;
 	}
 
 	/* if no -e flag read from command file descriptor */
@@ -811,9 +815,9 @@ static int cmdline(char	*cbuf)		/* uses eflag, eargc, cmdf */
 		if ((*++cbuf = inc) == '\\')		/* if it's escape */ 
 			*++cbuf = inc = getc(cmdf);	/* get next char */
 		else if (*cbuf == '\n')			/* end on newline */
-			return(*cbuf = '\0', 1);	/* cap the string */
+			return *cbuf = '\0', 1;	/* cap the string */
 
-	return(*++cbuf = '\0', -1);	/* end-of-file, no more chars */
+	return *++cbuf = '\0', -1;	/* end-of-file, no more chars */
 }
 
 /* expand an address at *cp... into expbuf, return ptr at following char */
@@ -829,10 +833,10 @@ static char *address(char *expbuf)		/* uses cp, linenum */
 		*expbuf++ = CEOF;	/* and the end-of-address mark (!) */
 		cp++;			/* go to next source character */
 		last_line_used = TRUE;
-		return(expbuf);		/* we're done */
+		return expbuf;		/* we're done */
 	}
 	if (*cp == '/')			/* start of regular-expression match */
-		return(recomp(expbuf, *cp++));	/* compile the RE */
+		return recomp(expbuf, *cp++);	/* compile the RE */
 
 	rcp = cp; lno = 0;		/* now handle a numeric address */
 	while(*rcp >= '0' && *rcp <= '9')	/* collect digits */
@@ -847,10 +851,10 @@ static char *address(char *expbuf)		/* uses cp, linenum */
 			die(TMLNR);	/*   abort with error message */
 		*expbuf++ = CEOF;	/* write the end-of-address marker */
 		cp = rcp;		/* point compile past the address */ 
-		return(expbuf);		/* we're done */
+		return expbuf;		/* we're done */
 	}
 
-	return(NULL);		/* no legal address was found */
+	return NULL;		/* no legal address was found */
 }
 
 /* accept multiline input from *cp..., discarding leading whitespace
@@ -864,11 +868,11 @@ static char *gettext(char* txp)		/* uses global cp */
 		if ((*txp = *p++) == '\\')	/* handle escapes */
 			*txp = *p++;
 		if (*txp == '\0')		/* we're at end of input */
-			return(cp = --p, ++txp);
+			return cp = --p, ++txp;
 		else if (*txp == '\n')		/* also SKIPWS after newline */
 			SKIPWS(p);
 	} while (txp++);		/* keep going till we find that nul */
-	return(txp);
+	return txp;
 }
 
 /* find the label matching *ptr, return NULL if none */
@@ -877,8 +881,8 @@ static label *search(label *ptr)		/* uses global lablst */
 	register label	*rp;
 	for(rp = lablst; rp < ptr; rp++)
 		if ((rp->name != NULL) && (strcmp(rp->name, ptr->name) == 0))
-			return(rp);
-	return(NULL);
+			return rp;
+	return NULL;
 }
 
 /* write label links into the compiled-command space */
@@ -920,7 +924,7 @@ static char *ycomp(char *ep, char delim)
 		if (*tp == '\\')
 			tp++;
 		if ((*tp == '\n') || (*tp == '\0'))
-			return(BAD);
+			return BAD;
 	}
 	tp++;		/* tp now points at first char of 'to' section */
 
@@ -938,11 +942,11 @@ static char *ycomp(char *ep, char delim)
 			tp++;
 		}
 		if ((ep[c] == delim) || (ep[c] == '\0'))
-			return(BAD);
+			return BAD;
 	}
 
 	if (*tp != delim)	/* 'to', 'from' parts have unequal lengths */
-		return(BAD);
+		return BAD;
 
 	cp = ++tp;			/* point compile ptr past translit */
 
@@ -950,7 +954,7 @@ static char *ycomp(char *ep, char delim)
 		if (ep[c] == 0)
 			ep[c] = c;
 
-	return(ep + 0x80);	/* return first free location past table end */
+	return ep + 0x80;	/* return first free location past table end */
 }
 
 /* sedcomp.c ends here */
