@@ -1,6 +1,6 @@
 /* sedexec.c -- axecute compiled form of stream editor commands
    Copyright (C) 1995-2003 Eric S. Raymond
-   Copyright (C) 2004-2013 Rene Rebe
+   Copyright (C) 2004-2021 Rene Rebe
 
    The single entry point of this module is the function execute(). It
 may take a string argument (the name of a file to be used as text)  or
@@ -26,7 +26,7 @@ extern sedcmd	cmds[];		/* hold compiled commands */
 extern long	linenum[];	/* numeric-addresses table */
 
 /* miscellaneous shared variables */
-extern int	nflag;		/* -n option flag */
+extern short	nflag;		/* -n option flag */
 extern int	eargc;		/* scratch copy of argument count */
 extern sedcmd	*pending;	/* ptr to command waiting to be executed */
 
@@ -37,7 +37,7 @@ extern int	last_line_used; /* last line address ($) used */
 #define MAXHOLD		MAXBUF	/* size of the hold space */
 #define GENSIZ		MAXBUF	/* maximum genbuf size */
 
-static char LTLMSG[]	= "sed: line too long\n";
+static const char LTLMSG[]	= "sed: line too long\n";
 
 static char	*spend;		/* current end-of-line-buffer pointer */
 static long	lnum = 0L;	/* current source line number */
@@ -57,8 +57,6 @@ static int	lastline;		/* do-line flag */
 static int	line_with_newline;	/* line had newline */
 static int	jump;			/* jump to cmd's link address if set */
 static int	delete;			/* delete command flag */
-static int	needs_advance;		/* needs inc after substitution */
-					/* ugly HACK - neds REWORK */
 
 /* tagged-pattern tracking */
 static char	*bracend[MAXTAGS];	/* tagged pattern start pointers */
@@ -197,35 +195,33 @@ static int selected(sedcmd *ipc)
 }
 
 /* match RE at expbuf against linebuf; if gf set, copy linebuf from genbuf */
-static int match(char *expbuf, int gf)	/* uses genbuf */
+static int _match(char *expbuf, int gf)	/* uses genbuf */
 {
 	char *p1, *p2, c;
 
-	if (gf)
+	if (!gf)
+	{
+		p1 = linebuf;
+		locs = NULL;
+	}
+	else
 	{
 		if (*expbuf)
-			return(FALSE);
-		p1 = linebuf; p2 = genbuf;
-		while ((*p1++ = *p2++));
-		if (needs_advance) {
+			return FALSE;
+		/* if the last match was zero length, continue to next */
+		if (loc2 - loc1 == 0) {
 			loc2++;
 		}
 		locs = p1 = loc2;
 	}
-	else
-	{
-		p1 = linebuf + needs_advance;
-		locs = FALSE;
-	}
-	needs_advance = 0;
-
+	
 	p2 = expbuf;
 	if (*p2++)
 	{
 		loc1 = p1;
-		if(*p2 == CCHR && p2[1] != *p1)	/* 1st char is wrong */
-			return(FALSE);		/*   so fail */
-		return(advance(p1, p2, NULL));	/* else try to match rest */
+		if (*p2 == CCHR && p2[1] != *p1)	/* 1st char is wrong */
+			return FALSE;		/*   so fail */
+		return advance(p1, p2, NULL);	/* else try to match rest */
 	}
 
 	/* quick check for 1st character if it's literal */
@@ -235,22 +231,33 @@ static int match(char *expbuf, int gf)	/* uses genbuf */
 		do {
 			if (*p1 != c)
 				continue;	/* scan the source string */
-			if (advance(p1, p2,NULL)) /* found it, match the rest */
-				return(loc1 = p1, 1);
-		} while
-			(*p1++);
-		return(FALSE);		/* didn't find that first char */
+			if (advance(p1, p2, NULL)) /* found it, match the rest */
+				return loc1 = p1, TRUE;
+		} while (*p1++);
+		return FALSE;		/* didn't find that first char */
 	}
 
 	/* else try for unanchored match of the pattern */
 	do {
 		if (advance(p1, p2, NULL))
-			return(loc1 = p1, 1);
-	} while
-		(*p1++);
+			return loc1 = p1, TRUE;
+	} while (*p1++);
 
 	/* if got here, didn't match either way */
-	return(FALSE);
+	return FALSE;
+}
+
+static int match(char *expbuf, int gf)	/* uses genbuf */
+{
+	const char *loc2i = loc2;
+	const int ret = _match(expbuf, gf);
+	
+	/* if last match was zero length, do not allow a follpwing zero match */
+	if (loc2i && loc1 == loc2i && loc2 - loc1 == 0) {
+		loc2++;
+		return _match(expbuf, gf);
+	}
+	return ret;
 }
 
 /* attempt to advance match pointer by one pattern element
@@ -270,31 +277,31 @@ static int advance(char* lp, char* ep, char** eob)
 		case CCHR:		/* literal character */
 			if (*ep++ == *lp++)	/* if chars are equal */
 				continue;	/* matched */
-			return(FALSE);		/* else return false */
+			return FALSE;		/* else return false */
 
 		case CDOT:		/* anything but newline */
 			if (*lp++)		/* first NUL is at EOL */
 				continue;	/* keep going if didn't find */
-			return(FALSE);		/* else return false */
+			return FALSE;		/* else return false */
 
 		case CNL:		/* start-of-line */
 		case CDOL:		/* end-of-line */
 			if (*lp == 0)		/* found that first NUL? */
 				continue;	/* yes, keep going */
-			return(FALSE);		/* else return false */
+			return FALSE;		/* else return false */
 
 		case CEOF:		/* end-of-address mark */
 			loc2 = lp;		/* set second loc */
-			return(TRUE);		/* return true */
+			return TRUE;		/* return true */
 
 		case CCL:		/* a closure */
-			c = *lp++ & 0177;
-			if (ep[c>>3] & bits(c & 07))	/* is char in set? */
+			c = *lp++;
+			if (ep[((unsigned char)c)>>3] & bits(c & 07))	/* is char in set? */
 			{
-				ep += 16;	/* then skip rest of bitmask */
+				ep += 32;	/* then skip rest of bitmask */
 				continue;	/*   and keep going */
 			}
-			return(FALSE);		/* else return false */
+			return FALSE;		/* else return false */
 
 		case CBRA:		/* start of tagged pattern */
 			brastart[(unsigned char)*ep++] = lp;	/* mark it */
@@ -304,7 +311,7 @@ static int advance(char* lp, char* ep, char** eob)
 			bcount = *ep;
 			if (eob) {
 				*eob = lp;
-				return (TRUE);
+				return TRUE;
 			}
 			else
 				bracend[(unsigned char)*ep++] = lp;    /* mark it */
@@ -319,7 +326,7 @@ static int advance(char* lp, char* ep, char** eob)
 				lp += ct;
 				continue;
 			}
-			return(FALSE);
+			return FALSE;
 
 		case CBRA|STAR:		/* \(...\)* */
 		{
@@ -343,7 +350,6 @@ static int advance(char* lp, char* ep, char** eob)
 				ep++;
 			ep+=2;
 
-			needs_advance = 1;
 			if (lp == curlp) /* 0 matches */
 				continue;
 			lp++; /* because the star handling decrements it */
@@ -359,10 +365,10 @@ static int advance(char* lp, char* ep, char** eob)
 			while(lp >= curlp)
 			{
 				if (advance(lp, ep, eob))
-					return(TRUE);
+					return TRUE;
 				lp -= ct;
 			}
-			return(FALSE);
+			return FALSE;
 
 		case CDOT|STAR:		/* match .* */
 			curlp = lp;		/* save closure start loc */
@@ -378,14 +384,12 @@ static int advance(char* lp, char* ep, char** eob)
 		case CCL|STAR:		/* match [...]* */
 			curlp = lp;		/* save closure start loc */
 			do {
-				c = *lp++ & 0x7F;	/* match any in set */
-			} while
-				(ep[c>>3] & bits(c & 07));
-			ep += 16;		/* skip past the set */
+				c = *lp++;	/* match any in set */
+			} while (ep[((unsigned char)c)>>3] & bits(c & 07));
+			ep += 32;		/* skip past the set */
 			goto star;		/* match followers */
 
 		star:		/* the recursion part of a * or + match */
-			needs_advance = 1;
 			if (--lp == curlp) {	/* 0 matches */
 				continue;
 			}
@@ -397,10 +401,9 @@ static int advance(char* lp, char* ep, char** eob)
 					if (*lp != c)
 						continue;
 					if (advance(lp, ep, eob))
-						return(TRUE);
-				} while
-				(lp-- > curlp);
-				return(FALSE);
+						return TRUE;
+				} while (lp-- > curlp);
+				return FALSE;
 			}
 
 			if (*ep == CBACK)
@@ -410,10 +413,9 @@ static int advance(char* lp, char* ep, char** eob)
 					if (*lp != c)
 						continue;
 					if (advance(lp, ep, eob))
-						return(TRUE);
-				} while
-					(lp-- > curlp);
-				return(FALSE);
+						return TRUE;
+				} while (lp-- > curlp);
+				return FALSE;
 			}
 #endif
 			/* match followers, try shorter match, if needed */
@@ -421,10 +423,9 @@ static int advance(char* lp, char* ep, char** eob)
 				if (lp == locs)
 					break;
 				if (advance(lp, ep, eob))
-					return(TRUE);
-			} while
-				(lp-- > curlp);
-			return(FALSE);
+					return TRUE;
+			} while (lp-- > curlp);
+			return FALSE;
 
 		default:
 			fprintf(stderr, "sed: internal RE error, %o\n", *--ep);
@@ -436,21 +437,18 @@ static int advance(char* lp, char* ep, char** eob)
    ipc:	ptr to s command struct */
 static int substitute(sedcmd *ipc)
 {
-	unsigned int n = 1;
+	unsigned int n = 0;
 	/* find a match */
-	/* the needs_advance code got a bit tricky - might needs a clean
-	   refactoring */
-	while (match(ipc->u.lhs, 0)) {
+	while (match(ipc->u.lhs, n /* use last loc2 for n>0 */)) {
 		/* nth 0 is implied 1 */
+		n++;
 		if (!ipc->nth || n == ipc->nth) {
 			dosub(ipc->rhs);		/* perform it once */
-			n++;				/* mark for return */
 			break;
 		}
-		needs_advance = n++;
 	}
-	if (n == 1)
-		return(FALSE);			/* command fails */
+	if (n == 0)
+		return FALSE;			/* command fails */
 
 	if (ipc->flags.global)			/* if global flag enabled */
 		do {				/* cycle through possibles */
@@ -460,7 +458,7 @@ static int substitute(sedcmd *ipc)
 			else				/* otherwise, */
 				break;			/* we're done */
 		} while (*loc2);
-	return(TRUE);				/* we succeeded */
+	return TRUE;				/* we succeeded */
 }
 
 /* generate substituted right-hand side (of s command)
@@ -474,6 +472,7 @@ static void dosub(char *rhsbuf)		/* uses linebuf, genbuf, spend */
 	lp = linebuf; sp = genbuf;
 	while (lp < loc1) *sp++ = *lp++;
 
+	/* substitute */
 	for (rp = rhsbuf; (c = *rp++); )
 	{
 		if (c & 0200 && (c & 0177) == '0')
@@ -491,8 +490,14 @@ static void dosub(char *rhsbuf)		/* uses linebuf, genbuf, spend */
 			fprintf(stderr, LTLMSG);
 
 	}
+
+	/* adjust location pointers and copy reminder */
 	lp = loc2;
-	loc2 = sp - genbuf + linebuf;
+	{
+		long len = loc2 - loc1;
+		loc2 = sp - genbuf + linebuf;
+		loc1 = loc2 - len;
+	}
 	while ((*sp++ = *lp++))
 		if (sp >= genbuf + MAXBUF)
 			fprintf(stderr, LTLMSG);
@@ -510,10 +515,10 @@ static char *place(char* asp, char* al1, char* al2)		/* uses genbuf */
 		if (asp >= genbuf + MAXBUF)
 			fprintf(stderr, LTLMSG);
 	}
-	return(asp);
+	return asp;
 }
 
-/* list the pattern space in  visually unambiguous form *p1... to fp
+/* list the pattern space in visually unambiguous form *p1... to fp
    p1: the source
    fp: output stream to write to */
 static void listto(char *p1, FILE *fp)
@@ -569,7 +574,6 @@ static void command(sedcmd *ipc)
 	register char	*p1, *p2;
 	char		*execp;
 
-	needs_advance = 0;
 	switch(ipc->command)
 	{
 	case ACMD:		/* append */
@@ -588,7 +592,7 @@ static void command(sedcmd *ipc)
 		break;
 
 	case DCMD:		/* delete pattern space */
-		delete++;
+		delete = TRUE;
 		break;
 
 	case CDCMD:		/* delete a line in hold space */
@@ -786,11 +790,11 @@ static char *sed_getline(char *buf, int max)
 		  }
 		}
 
-		return(buf);		/* return ptr to terminating null */ 
+		return buf;		/* return ptr to terminating null */ 
 	}
 	else
 	{
-		return(BAD);
+		return BAD;
 	}
 }
 
